import unittest
import math
import numpy as np
from bagua.autotune import (
    BayesianOptimizer,
    IntParam,
    FloatParam,
    autotune_system_hyperparameters,
)


class TestBayesianOptimizer(unittest.TestCase):
    def test_autotune_system_hyperparameters(self):
        autotune_system_hyperparameters(["localhost"], 8, 8123)

    def test_bayesian_optimization(self):
        def f(x, y):
            x /= 10000.0
            y /= 10000.0
            return np.sin(5 * x) * (1 - np.tanh(y ** 2)) + np.random.randn() * 0.1

        optim = BayesianOptimizer(
            {
                "x": IntParam(
                    val=1,
                    space_dimension=[-20000, 20000],
                ),
                "y": IntParam(
                    val=1,
                    space_dimension=[-20000, 20000],
                ),
            },
        )

        param_score_list = []

        # Tune
        for i in range(100):
            d = optim.ask()
            score = f(d["x"], d["y"])
            optim.tell(d, score)
            param_score_list.append([d, score])

        # Test
        best_score = sorted(param_score_list, key=lambda p_s: -p_s[1])[0][1]
        self.assertTrue(best_score > 0.8)


if __name__ == "__main__":
    unittest.main()



import time
import argparse

import cupy
import torch
import torch.distributed as dist

from bagua.torch_api.distributed import init_sub_communicator
from bagua.torch_api.utils import cupy2torch

# import sys
# sys.path.append("../")
# sys.path.append("../../")
# from tests.nccl.communicators import init_sub_communicator, torch2cupy, cupy2torch


def run(args):
    """Distributed Synchronous SGD Example"""
    device = torch.device(
        "cuda:{}".format(args.gpu_id)
        if torch.cuda.is_available() and not args.no_cuda
        else "cpu"
    )
    print("device={}".format(device))
    torch.cuda.set_device(args.gpu_id)
    torch.manual_seed(args.rank)

    round = 10
    comm_ranks = [1, 2, 3]
    sg_ranks = [0, 1, 2]  # rank in subgroup
    sg_size = len(sg_ranks)

    comm = None
    if args.rank in comm_ranks:
        index = comm_ranks.index(args.rank)
        sg_rank = sg_ranks[index]
        comm = init_sub_communicator("nccl", sg_rank, sg_size)
        print("subgroup rank: {}, size: {}".format(sg_rank, sg_size))

    send_size = 100

    for i in range(round):
        if args.rank in comm_ranks:
            torch_tensor = torch.full(
                (sg_size, send_size), args.rank, dtype=torch.float64
            )
            # print("torch tensor shape = {}".format(torch_tensor.shape))
            # cupy_tensor = torch2cupy(torch_tensor).reshape((sg_size-1, send_size))
            sendbuf = cupy.asarray(torch_tensor.numpy())
            # print("sendbuf shape = {}".format(sendbuf.shape))
            recvbuf = cupy.zeros((sg_size, send_size), dtype=cupy.float64)
            comm_start = time.time()
            comm.scatter_all(sendbuf, recvbuf, send_size)
            comm_end = time.time()
            print("communication cost {} s".format(comm_end - comm_start))
            torch_tensor = cupy2torch(recvbuf)
            print(
                "tensor after communication {} = {}".format(
                    torch_tensor.shape, torch_tensor
                )
            )
        dist.barrier()
        print(">>>>>>>>>>> round {} finishes".format(i))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--backend", type=str, default="nccl", help="Name of the backend to use."
    )
    parser.add_argument(
        "-i",
        "--init-method",
        type=str,
        default="env://",
        help="URL specifying how to initialize the package.",
    )
    parser.add_argument(
        "-s",
        "--world-size",
        type=int,
        default=1,
        help="Number of processes participating in the job.",
    )
    parser.add_argument(
        "-r", "--rank", type=int, default=0, help="Rank of the current process."
    )
    parser.add_argument("--no-cuda", action="store_true")
    parser.add_argument("--gpu-id", type=int, default=1)
    args = parser.parse_args()
    print(args)

    if args.world_size > 1:
        dist.init_process_group(
            backend=args.backend,
            init_method=args.init_method,
            world_size=args.world_size,
            rank=args.rank,
        )

    run(args)


if __name__ == "__main__":
    main()



_global_state = None


def _set_global_state(global_state):
    global _global_state
    _global_state = global_state


def _get_global_state():
    global _global_state
    return _global_state


def is_initialized():
    """
    Checking if bagua global communication state has been initialized
    """
    global _global_state
    return _global_state is not None



#!/usr/bin/env python3

import torch
import bagua_ext


def test_flatten():
    t1 = torch.tensor([1, 2, 3])
    t2 = torch.tensor([4, 5, 6])
    t = bagua_ext.flatten([t1, t2])
    assert (t - torch.tensor([1, 2, 3, 4, 5, 6])).abs().sum() == 0


def test_unflatten():
    t1 = torch.tensor([1, 2, 3])
    t2 = torch.tensor([4, 5, 6])
    t3, t4 = bagua_ext.unflatten(torch.tensor([3, 2, 1, 6, 5, 4]), [t1, t2])
    assert torch.allclose(t3, torch.tensor([3, 2, 1]))
    assert torch.allclose(t4, torch.tensor([6, 5, 4]))


if __name__ == "__main__":
    test_flatten()
    test_unflatten()



from collections import OrderedDict
import torch.distributed as dist
import torch
import logging
import numpy as np

LOGGER = logging.getLogger(__name__)

flatten = torch._utils._flatten_dense_tensors
unflatten = torch._utils._unflatten_dense_tensors


def apply_flattened_call(bucket, call, extra_args=None):
    coalesced = flatten(bucket)

    if extra_args is not None:
        call(coalesced, *extra_args)
    else:
        call(coalesced)

    if call is dist.all_reduce:
        coalesced /= dist.get_world_size()

    for buf, synced in zip(bucket, unflatten(coalesced, bucket)):
        buf.copy_(synced)


def _group_by_tensor_type(tensors):
    buckets = OrderedDict()
    for tensor in tensors:
        tp = tensor.type()
        if tp not in buckets:
            buckets[tp] = []
        buckets[tp].append(tensor)
    return buckets


def apply_flattened_call_all(tensors, call):
    """
    Apply call on a list of tensors.
    """

    grouped_tensors = _group_by_tensor_type(tensors)
    for tensors in grouped_tensors.values():
        apply_flattened_call(tensors, call)


def align_size(size, align):
    return int((size + align - 1) / align) * align


def check_contiguous(tensors):
    data_ptr = None
    for t in tensors:
        if data_ptr is not None and t.data_ptr() != data_ptr:
            return False
        data_ptr = t.data_ptr() + t.numel() * t.element_size()
    return True


def _get_params_flattened_aligned_size(params, align_bytes):
    assert align_bytes == 1 or (
        align_bytes % params[0].element_size() == 0
    ), "align bytes must be multiples of element size"

    sizes = [p.numel() for p in params]

    total_size = sum(sizes)
    aligned_total_size = (
        align_size(total_size * params[0].element_size(), align_bytes)
        // params[0].element_size()
    )

    # padding to the last param
    sizes[-1] += aligned_total_size - total_size

    for p, sz in zip(params, sizes):
        p.allocated_size = sz

    return aligned_total_size


def flatten_module_params(params_list, align_bytes: int):
    if len(params_list) == 0:
        return

    if not isinstance(params_list[0], list):
        params_list = [params_list]

    total_size = 0
    for params in params_list:
        total_size += _get_params_flattened_aligned_size(params, align_bytes)

    logging.debug(
        f"flatten {str(params_list[0][0].dtype).partition('.')[-1]} params aligned to {align_bytes} bytes, total numels: {total_size}"
    )

    flatten_weights_tensor = torch.zeros(total_size, dtype=params_list[0][0].dtype).to(
        params_list[0][0].device
    )
    flatten_grads_tensor = torch.zeros(total_size, dtype=params_list[0][0].dtype).to(
        params_list[0][0].device
    )

    flatten_weights_storage = flatten_weights_tensor.storage()
    flatten_grads_storage = flatten_grads_tensor.storage()

    def set_storage(param, weight_storage, grad_storage, storage_offset):
        with torch.no_grad():
            z = torch.zeros_like(param.data)
            z.set_(weight_storage, storage_offset, param.shape)
            param.data = z

            t = torch.zeros_like(param.data)
            t.set_(grad_storage, storage_offset, param.shape)
            param.grad = t

    offset = 0
    for params in params_list:
        for p in params:
            # copy data
            flatten_weights_tensor[offset : offset + p.numel()] = p.data.reshape(-1)

            if p.grad is not None:
                flatten_grads_tensor[offset : offset + p.numel()] = p.grad.data.reshape(
                    -1
                )
            else:
                logging.debug(f"grad is none, {offset}")

            # flatten
            set_storage(p, flatten_weights_storage, flatten_grads_storage, offset)
            offset += p.allocated_size
            logging.debug(f"flatten param done {offset}")

    # # check
    for params in params_list:
        weight_tensors = [p.data for p in params]
        grad_tensors = [p.grad.data for p in params]

        assert check_contiguous(weight_tensors)
        assert check_contiguous(grad_tensors)

    return new_param(flatten_weights_tensor, flatten_grads_tensor)


def collocate_params(params):
    """
    `tensors` share the same storage
    """
    if len(params) == 1:
        return params[0]

    logging.debug(f"fuse {len(params)} params")

    sorted_params = sorted(params, key=lambda x: x.storage_offset())

    start = None
    offset = 0
    for p in sorted_params:
        weight = p.data
        grad = p.grad.data

        assert (
            weight.storage_offset() == grad.storage_offset()
        ), "collocated weights and grads must have consistent storage offset"

        if start is None:
            start = offset = weight.storage_offset()
        else:
            assert (
                offset == weight.storage_offset()
            ), "params collocated must be contiguous"

        offset += (
            p.bagua_tensor.num_elem_allocated()
            if hasattr(p, "bagua_tensor")
            else p.numel()
        )

    with torch.no_grad():
        weight_tensor = torch.zeros(offset - start, dtype=params[0].dtype).to(
            params[0].device
        )
        weight_tensor.set_(params[0].data.storage(), start, weight_tensor.shape)

        grad_tensor = torch.zeros(offset - start, dtype=params[0].dtype).to(
            params[0].device
        )
        grad_tensor.set_(params[0].grad.data.storage(), start, grad_tensor.shape)

        return new_param(weight_tensor, grad_tensor)


def new_param(weight, grad):
    with torch.no_grad():
        p = torch.nn.Parameter(weight, requires_grad=False)
        p.grad = grad
        return p


def to_bagua_datatype(datatype):
    if datatype == torch.float32:
        return "f32"
    elif datatype == torch.float16:
        return "f16"
    elif datatype == torch.uint8:
        return "u8"
    elif datatype == torch.long:
        return "i64"
    else:
        raise ValueError(f"unsupported data type {datatype}.")


def to_bagua_reduce_op(torch_reduce_op):
    if torch_reduce_op == dist.ReduceOp.SUM:
        return 0
    elif torch_reduce_op == dist.ReduceOp.PRODUCT:
        return 1
    elif torch_reduce_op == dist.ReduceOp.MIN:
        return 2
    elif torch_reduce_op == dist.ReduceOp.MAX:
        return 3
    elif torch_reduce_op == dist.ReduceOp.BOR:
        return 7
    elif torch_reduce_op == dist.ReduceOp.BAND:
        return 8
    elif torch_reduce_op == dist.ReduceOp.BXOR:
        return 9
    else:
        raise ValueError(f"unsupported reduce op {torch_reduce_op}.")


def average_by_removing_extreme_values(raw_score_list):
    score_list = np.asarray(raw_score_list)

    # weed out warm up data
    score_list = score_list[len(score_list) // 3 :]

    def weed_out_outliers(X):
        mean = np.mean(X)
        std = np.std(X)
        distance_from_mean = abs(X - mean)
        max_deivations = 1
        not_outlier = distance_from_mean < max_deivations * std
        not_outliers = X[not_outlier]

        if len(not_outliers) == 0:
            return X

        return not_outliers

    score_list = weed_out_outliers(score_list)

    # Repeat up to ten times
    for i in range(10):
        if np.std(score_list) < np.mean(score_list):
            break
        score_list = weed_out_outliers(score_list)

    # score = np.mean(score_list) # TODO: @shjwudp check whether these are still needed
    # std = np.std(score_list)

    return np.mean(score_list), np.std(score_list), score_list.tolist()



import enum
from typing import List
import sys
if sys.version_info >= (3, 9):
    from typing import TypedDict  # pytype: disable=not-supported-yet
else:
    from typing_extensions import TypedDict  # pytype: disable=not-supported-yet
from pydantic import BaseModel


class TensorDtype(str, enum.Enum):
    F32 = "f32"
    F16 = "f16"
    U8 = "u8"


class TensorDeclaration(TypedDict):
    name: str
    num_elements: int
    dtype: TensorDtype


class BaguaHyperparameter(BaseModel):
    """
    Structured all bagua hyperparameters
    """

    buckets: List[List[TensorDeclaration]] = []
    is_hierarchical_reduce: bool = False

    def update(self, param_dict: dict):
        tmp = self.dict()
        tmp.update(param_dict)
        self.parse_obj(tmp)
        return self



import time
import json
import logging
import requests
import socket
import threading
from flask import request, Flask
from werkzeug.middleware.dispatcher import DispatcherMiddleware
from prometheus_client import make_wsgi_app
from werkzeug.serving import make_server


def wait_for_port(port, host="localhost", timeout=5.0):
    """Wait until a port starts accepting TCP connections.
    Args:
        port (int): Port number.
        host (str): Host address on which the port should exist.
        timeout (float): In seconds. How long to wait before raising errors.
    Raises:
        TimeoutError: The port isn't accepting connection after time specified in `timeout`.
    """
    start_time = time.perf_counter()
    while True:
        try:
            with socket.create_connection((host, port), timeout=timeout):
                break
        except OSError as ex:
            time.sleep(0.01)
            if time.perf_counter() - start_time >= timeout:
                raise TimeoutError(
                    "Waited too long for the port {} on host {} to start accepting "
                    "connections.".format(port, host)
                ) from ex


class ServerThread(threading.Thread):
    def __init__(self, master_addr, master_port, app):
        threading.Thread.__init__(self)
        self.srv = make_server(master_addr, master_port, app)
        self.ctx = app.app_context()
        self.ctx.push()

    def run(self):
        self.srv.serve_forever()

    def shutdown(self):
        self.srv.shutdown()


def pick_n_free_ports(n: int):
    socks = []
    for i in range(n):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(("localhost", 0))
        socks.append(sock)

    n_free_ports = [sock.getsockname()[1] for sock in socks]
    for sock in socks:
        sock.close()

    return n_free_ports


def setup_app(slots: list, server_addr: str, server_port: int):
    # run http server
    app = Flask(__name__)
    app.use_reloader = False  # type: ignore

    @app.route("/get_fuselib_server_addr", methods=["POST"])
    def get_fuselib_server_addr():
        req: dict = request.get_json(force=True)
        hostname: str = req["hostname"]
        rank: int = req["rank"]

        slots[rank] = hostname

        return json.dumps({"server_addr": server_addr, "server_port": server_port})

    # Add prometheus wsgi middleware to route /metrics requests
    app.wsgi_app = DispatcherMiddleware(app.wsgi_app, {"/metrics": make_wsgi_app()})  # type: ignore

    return app


def generate_and_broadcast_server_addr(
    master_addr,
    master_port,
    world_size: int,
    my_rank: int,
    proxies={
        "http": None,
        "https": None,
    },
):
    # import torch
    # torch.distributed.init_process_group(backend="gloo", init_method="tcp://{}:{}".format(master_addr, master_port), world_size=world_size, rank=my_rank)
    # if my_rank == 0:
    #     objects = [pick_n_free_ports(1)[0]]
    # else:
    #     objects = [None]
    # dist.broadcast_object_list(objects, src=0)
    # dist.destroy_process_group()

    # server_port = broadcast_objects[0]

    # return (master_addr, server_port)

    slots = [None] * world_size
    if my_rank == 0:
        server_addr = master_addr
        server_port = pick_n_free_ports(1)[0]
        slots[my_rank] = socket.gethostname()  # type: ignore

        app = setup_app(slots, server_addr, server_port)
        server = ServerThread("0.0.0.0", master_port, app)
        server.start()

        while True:
            n_empty_slot = len([x for x in slots if x is None])
            if n_empty_slot == 0:
                break

            time.sleep(1)

        server.shutdown()
    else:
        # Wait service discovery service ready
        timeout = time.time() + 60  # 60s timeout
        rsp = None
        wait_for_port(host=master_addr, port=master_port)
        while time.time() < timeout:
            try:
                with requests.session() as sess:
                    rsp = sess.post(
                        "http://{}:{}/get_fuselib_server_addr".format(
                            master_addr, master_port
                        ),
                        json={"rank": my_rank, "hostname": socket.gethostname()},
                        proxies=proxies,
                        timeout=timeout,
                    )
            except requests.exceptions.ConnectionError as ex:
                logging.info(ex)

            if rsp and rsp.status_code == 200:
                break
            time.sleep(1)
        if rsp is None or rsp.status_code != 200:
            raise RuntimeError("Waiting for service discovery service start timeout")

        server_addr = json.loads(rsp.content)["server_addr"]
        server_port = json.loads(rsp.content)["server_port"]

    return (master_addr, server_port, slots)



#!/usr/bin/env python3

import torch
from utils import qsgd_compress, qsgd_decompress
import numpy as np
import cupy


def test_qsgd_compression():
    t = torch.tensor([0.1, 0.5, 1.5, -0.5, 2.4]).cuda()
    norm, sign, cupy_compressed_ints = qsgd_compress(t, pnorm=2.0)
    assert norm.item() == 2.9189038276672363
    assert sign.item() == 232
    assert (
        cupy.asnumpy(cupy_compressed_ints) - np.array([9, 44, 132, 44, 211]) <= 1
    ).all()


def test_qsgd_decompression():
    t = torch.tensor([0.1, 0.5, 1.5, -0.5, 2.4]).cuda()
    norm, sign, cupy_compressed_ints = qsgd_compress(t, pnorm=2.0)
    t_decompressed = qsgd_decompress(norm, sign, cupy_compressed_ints)
    assert ((t_decompressed - t).abs() <= 1 / 256 * norm).all()


if __name__ == "__main__":
    cupy.random.seed(42)
    torch.manual_seed(42)
    cupy_stream = cupy.cuda.ExternalStream(torch.cuda.current_stream().cuda_stream)
    cupy_stream.use()
    test_qsgd_compression()
    test_qsgd_decompression()



"""

Bagua_ is a communication library
developed by Kuaishou Technology and DS3 Lab for deep learning.

See tutorials_ for Bagua's rationale and benchmark.

.. _Bagua: https://github.com/BaguaSys/bagua
.. _tutorials: https://baguasys.github.io/tutorials/
"""

import bagua_core  # noqa: F401
from .version import __version__  # noqa: F401



#!/usr/bin/env python3
from typing import Optional

from bagua.torch_api.globals import _get_global_state
import torch
import bagua_core as B
import gorilla


@gorilla.patches(torch.Tensor, filter=lambda name, obj: "bagua" in name)
class BaguaTensor:
    """
    This class patch torch.Tensor with additional methods.
    """

    def _bagua_sanity_check(self):
        assert self._bagua_backend_tensor.data_ptr() == self.data_ptr()
        assert self._bagua_backend_tensor.num_elements() == self.numel()
        assert self._bagua_backend_tensor.num_elements_allocated() == self.numel()

    def is_bagua_tensor(self) -> bool:
        return hasattr(self, "_bagua_backend_tensor")

    def ensure_bagua_tensor(self, name: Optional[str] = None):
        """
        Convert a PyTorch tensor or parameter to Bagua tensor inplace and return it.
        A Bagua tensor is required to use Bagua's communication algorithms.

        Args:
            name: the unique name of the tensor

        Returns:
            The original tensor with Bagua tensor attributes initialized.
        """
        if self.is_bagua_tensor():
            if name is not None:
                assert (
                    self.bagua_tensor_name == name
                ), "assigning a different name to existing bagua tensor is forbidden"
            return
        self.bagua_tensor_name = name if name is not None else ""
        self._bagua_backend_tensor = B.BaguaTensorPy(
            name=self.bagua_tensor_name,
            torch_tensor=self,
        )
        self._bagua_sanity_check()
        self._bagua_backend = _get_global_state().get_backend()
        self._bagua_ready_event = torch.cuda.Event()
        self._bagua_bucket = None
        return self

    def to_bagua_tensor(self, name: Optional[str] = None):
        """
        Create a new Bagua tensor from a PyTorch tensor or parameter and return it.
        The original tensor is not changed. A Bagua tensor is required to use
        Bagua's communication algorithms.

        Args:
            name: the unique name of the tensor

        Returns:
            The new Bagua tensor sharing the same storage with the original tensor.
        """
        new_tensor = torch.Tensor(cdata=self._cdata)
        return new_tensor.ensure_bagua_tensor(name)

    def bagua_backend_tensor(self) -> B.BaguaTensorPy:
        """
        Returns:
            The raw Bagua backend tensor.
        """
        return self._bagua_backend_tensor

    def bagua_ensure_grad(self) -> torch.Tensor:
        """
        Return the gradient of current parameter. Create a zero gradient tensor
        if not exist.
        """
        if hasattr(self, "grad") and self.grad is not None:
            return self.grad
        elif isinstance(self, torch.nn.Parameter):
            with torch.no_grad():
                t = torch.zeros_like(self.data)
                self.grad = t
            return self.grad
        else:
            raise NotImplementedError

    def bagua_mark_communication_ready(self):
        """
        Mark a Bagua tensor ready for scheduled operations execution.
        """
        torch.cuda.current_stream().record_event(self._bagua_ready_event)
        self._bagua_backend.mark_communication_ready(
            self._bagua_backend_tensor,
            self._bagua_ready_event.cuda_event,
        )

    def bagua_mark_communication_ready_without_synchronization(self):
        """
        Mark a Bagua tensor ready immediately, without CUDA event synchronization.
        """
        self._bagua_backend.mark_communication_ready(
            self._bagua_backend_tensor,
            0,
        )

    def bagua_set_storage(self, storage: torch.Storage, storage_offset: int = 0):
        """
        Sets the underlying storage using an existing torch.Storage.

        Args:
            storage: the storage to use
            storage_offset: the offset in the storage
        """
        with torch.no_grad():
            self.set_(storage, storage_offset, self.shape)


_base = gorilla._get_base(BaguaTensor)
_decorator_data = gorilla.get_decorator_data(_base)
for patch in _decorator_data.patches:
    gorilla.apply(patch)



import torch
from bagua.torch_api.fuse_optimizer import FusedOptimizer
import logging


logging.basicConfig(level=logging.DEBUG)


def _get_aligned_flatten_tensor(tensors):
    flatten = torch._utils._flatten_dense_tensors
    flatten_tensor = flatten(tensors)
    t = torch.zeros(
        int((flatten_tensor.numel() / 8 + 1) * 8), dtype=flatten_tensor.dtype
    ).to(flatten_tensor.device)
    t[0 : flatten_tensor.numel()] = flatten_tensor
    return t


def _init_params():
    torch.manual_seed(49)
    t1 = torch.rand((10, 10), requires_grad=True)
    t2 = torch.rand((10, 10), requires_grad=True)
    t3 = torch.add(t1, t2)

    t4 = torch.rand((10, 10), requires_grad=True)
    t5 = torch.mul(t3, t4)

    t6 = torch.rand((10, 10), requires_grad=True)
    t7 = torch.rand((10, 10), requires_grad=True)
    t8 = torch.mul(t6, t7)

    t9 = torch.add(t5, t8)

    loss = t9.sum()
    loss.backward()

    return [t1, t2, t4, t6, t7]


def _flatten_params(params):

    param_tensors = list(map(lambda x: x.data, params))

    flatten_tensor = _get_aligned_flatten_tensor(param_tensors)
    flatten_storage = flatten_tensor.storage()
    offset = 0

    for p in params:
        with torch.no_grad():
            p.set_(flatten_storage, offset, p.shape)

        offset += p.numel()
    return flatten_tensor


def _flatten_params_grad(params):
    grad_tensors = list(map(lambda x: x.grad.data, params))

    flatten_tensor = _get_aligned_flatten_tensor(grad_tensors)
    flatten_storage = flatten_tensor.storage()
    offset = 0

    for p in params:
        with torch.no_grad():
            t = torch.zeros_like(p.data)
            t.set_(flatten_storage, offset, p.shape)
            p.grad = t

        offset += p.numel()
    return flatten_tensor


def test_unfused():
    params = _init_params()
    params_group = [
        {"params": params[0:-2], "weight_decay": 0.0},
        {"params": params[-2:], "weight_decay": 0.8},
    ]

    optimizer = torch.optim.SGD(params_group, lr=0.001)

    optimizer.step()

    flattened = _flatten_params(params)
    logging.info(f"flattened param norm: {torch.norm(flattened)}")


def test_unflattened():
    params = _init_params()
    params_group = [
        {"params": params[0:-2], "weight_decay": 0.0},
        {"params": params[-2:], "weight_decay": 0.8},
    ]

    optimizer = torch.optim.SGD(params_group, lr=0.001)
    fused_optimzer = FusedOptimizer(optimizer)

    fused_optimzer.step()

    flattened = _flatten_params(params)
    logging.info(f"flattened param norm: {torch.norm(flattened)}")


def test_flatten_grads():
    params = _init_params()
    params_group = [
        {"params": params[0:-2], "weight_decay": 0.0},
        {"params": params[-2:], "weight_decay": 0.8},
    ]

    _flatten_params_grad(params)
    optimizer = torch.optim.SGD(params_group, lr=0.001)
    fused_optimzer = FusedOptimizer(optimizer)

    fused_optimzer.step()
    flattened = _flatten_params(params)
    logging.info(f"flattened param norm: {torch.norm(flattened)}")


def test_flatten_weights():
    params = _init_params()
    params_group = [
        {"params": params[0:-2], "weight_decay": 0.0},
        {"params": params[-2:], "weight_decay": 0.8},
    ]

    _flatten_params(params)
    optimizer = torch.optim.SGD(params_group, lr=0.001)
    fused_optimzer = FusedOptimizer(optimizer)

    fused_optimzer.step()
    flattened = _flatten_params(params)
    logging.info(f"flattened param norm: {torch.norm(flattened)}")


def test_partial_flatten():
    params = _init_params()
    params_group = [
        {"params": params[0:-2], "weight_decay": 0.0},
        {"params": params[-2:], "weight_decay": 0.8},
    ]

    _flatten_params(params[1:-2])
    _flatten_params_grad(params[1:-2])

    optimizer = torch.optim.SGD(params_group, lr=0.001)
    fused_optimzer = FusedOptimizer(optimizer)

    fused_optimzer.step()
    flattened = _flatten_params(params)
    logging.info(f"flattened param norm: {torch.norm(flattened)}")


def test_all_flatten():
    params = _init_params()
    params_group = [
        {"params": params[0:-2], "weight_decay": 0.0},
        {"params": params[-2:], "weight_decay": 0.8},
    ]

    _flatten_params(params)
    _flatten_params_grad(params)

    optimizer = torch.optim.SGD(params_group, lr=0.001)
    fused_optimzer = FusedOptimizer(optimizer)

    fused_optimzer.step()
    flattened = _flatten_params(params)
    logging.info(f"flattened param norm: {torch.norm(flattened)}")


if __name__ == "__main__":
    test_unfused()
    test_unflattened()
    test_flatten_grads()
    test_flatten_weights()
    test_partial_flatten()
    test_all_flatten()



import os


def get_world_size():
    """
    Get the number of processes in the current process group.

    Returns:
        The world size of the process group.
    """
    return int(os.environ.get("WORLD_SIZE", 1))


def get_rank():
    """
    Get the rank of current process group.

    Rank is a unique identifier assigned to each process within a distributed
    process group. They are always consecutive integers ranging from 0 to
    ``world_size``.

    Returns:
        The rank of the process group.
    """
    return int(os.environ.get("RANK", 0))


def get_local_rank():
    """
    Get the rank of current node.

    Local rank is a unique identifier assigned to each process within a node.
    They are always consecutive integers ranging from 0 to ``local_size``.

    Returns:
        The local rank of the node.
    """
    return int(os.environ.get("LOCAL_RANK", 0))


def get_local_size():
    """
    Get the number of processes in the node.

    Returns:
        The local size of the node.
    """
    return int(os.environ.get("LOCAL_WORLD_SIZE", 1))


def get_default_bucket_size() -> int:
    """Get default communication bucket byte size.

    Returns:
        int: default bucket size
    """
    return int(os.environ.get("BAGUA_DEFAULT_BUCKET_SIZE", 10 * 1024 ** 2))


def get_autotune_level() -> int:
    """Get the atuotune level.

    Returns:
        int: The autotune level.
    """
    return int(os.environ.get("BAGUA_AUTOTUNE", 0))


def get_master_addr():
    return os.environ.get("MASTER_ADDR", "127.0.0.1")


def get_bagua_service_port():
    return int(os.environ.get("BAGUA_SERVICE_PORT", -1))


def is_report_metrics_switch_on():
    """
    Whether bagua report switch is on or not.
    """
    return int(os.environ.get("BAGUA_REPORT_METRICS", 0)) == 1



import copy
import csv
import requests
import os
import time
import threading
import json
import logging
import math
import multiprocessing
from flask import Flask, request, session
from werkzeug.middleware.dispatcher import DispatcherMiddleware
import prometheus_client
from prometheus_client import make_wsgi_app
from bagua.autotune import BayesianOptimizer, IntParam, BoolParam
from bagua.bagua_define import (
    TensorDtype,
    TensorDeclaration,
    BaguaHyperparameter,
)
import numpy as np
from typing import Dict, List, Tuple
from collections import OrderedDict


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, TensorDtype):
            return obj.value
        else:
            return super(NpEncoder, self).default(obj)


# TODO: There is only fusion logic, should the logic of splitting tensor be added?
def split_bucket_by_bucket_size(
    tensor_list: List[TensorDeclaration],
    bucket_size: int,
    param_group_info: Dict[str, int] = {},
):
    dtype_unit_size = {
        TensorDtype.F32.value: 4,
        TensorDtype.F16.value: 2,
        TensorDtype.U8.value: 1,
    }

    buckets = []
    tmp_bucket = []
    tmp_bucket_size = 0
    for (dtype, unit_size) in sorted(dtype_unit_size.items()):
        for tensor_declar in [x for x in tensor_list if x["dtype"] == dtype]:
            tensor_size = tensor_declar["num_elements"] * unit_size
            tmp_bucket_size += tensor_size
            tmp_bucket.append(tensor_declar)
            if tmp_bucket_size >= bucket_size:
                buckets.append(tmp_bucket)
                tmp_bucket = []
                tmp_bucket_size = 0

    if len(tmp_bucket) != 0:
        buckets.append(tmp_bucket)

    # group params in bucket
    for i, _ in enumerate(buckets):
        buckets[i] = sorted(
            buckets[i], key=lambda p: param_group_info.get(p["name"], -1)
        )

    return buckets


try:
    rank0_hyperparameters = prometheus_client.Info(
        "rank0_hyperparameters", "The newly reported rank0 hyperparameter."
    )
    rank0_train_iter = prometheus_client.Gauge(
        "rank0_train_iter", "The iteration when parameters reporting."
    )
    rank0_denoised_iter_per_seconds = prometheus_client.Gauge(
        "rank0_denoised_iter_per_seconds",
        "Data used for hyperparameters evaluation.",
    )
except ValueError:
    pass


def record_autotune_log(
    autotune_log_filepath: str, autotune_hp: dict, train_iter: int, score: float
):
    with open(autotune_log_filepath, "a") as autotune_log:
        csv_writer = csv.DictWriter(
            autotune_log,
            fieldnames=sorted(["train_iter", "score"] + list(autotune_hp.keys())),
        )
        first_line = open(autotune_log_filepath).readline()
        if not first_line:
            csv_writer.writeheader()

        cols = copy.deepcopy(autotune_hp)
        cols.update(
            {
                "train_iter": train_iter,
                "score": score,
            }
        )
        cols = OrderedDict(cols)
        logging.info("cols={}".format(cols))
        csv_writer.writerow(cols)


class AutotuneService:
    def __init__(
        self,
        world_size,
        autotune_level=0,
        max_samples=60,
        sampling_confidence_time_s=5,
        warmup_time_s=30,
        autotune_log_filepath="/tmp/bagua_autotune.log",
        default_bucket_size=10 * 1024 ** 2,
    ):
        self.autotune_level = autotune_level
        self.world_size = world_size
        self.max_samples = max_samples
        self.sampling_confidence_time_s = sampling_confidence_time_s
        self.warmup_time_s = warmup_time_s
        self.warmup_flag = False
        self.is_initialized = False
        self.autotune_log_filepath = autotune_log_filepath
        if self.autotune_level >= 1:
            try:
                os.remove(self.autotune_log_filepath)
            except OSError:
                pass

        self.bayesian_optimizer = BayesianOptimizer(
            {
                "bucket_size_2p": IntParam(  # bucket_size = 2 ^ bucket_size_2p
                    val=13,
                    space_dimension=(  # 0 ~ 2GB
                        10,
                        31,
                    ),
                ),
                "is_hierarchical_reduce": BoolParam(False),
            },
            n_initial_points=10,
        )
        self.hyperparameters_and_score_list: list = []
        self.sampling_counter: int = 0  # 采样计数器
        self.check_board = [-1] * self.world_size
        self.last_time_the_hyperparameters_was_granted = (
            time.time()
        )  # 记录上次超参授予bagua进程的时刻，用于衡量采样时间

        self.ask_hyperparameters_mutex = threading.Lock()
        self.recommended_hyperparameters: BaguaHyperparameter = BaguaHyperparameter()
        self.recommended_from_iter = 0
        self.param_group_info: Dict[str, int] = {}
        self.default_bucket_size: int = default_bucket_size
        self.sess_bucket_size: int = self.default_bucket_size

        # metrics to autotune
        self.metrics_mutex = threading.Lock()
        rank0_hyperparameters.info(self.recommended_hyperparameters.dict())
        rank0_train_iter.set(-1)
        rank0_denoised_iter_per_seconds.set(-1)

    def tell_and_ask_bayesian_recommended_hyperparameters(
        self, bagua_hp, score, train_iter, bucket_size_2p
    ) -> Tuple[BaguaHyperparameter, Dict]:
        autotune_hp = {
            "bucket_size_2p": bucket_size_2p,
            "is_hierarchical_reduce": bagua_hp.is_hierarchical_reduce,
        }
        self.bayesian_optimizer.tell(autotune_hp, score)
        recommended_autotune_hp = self.bayesian_optimizer.ask()
        recommended_bucket_size = 2 ** recommended_autotune_hp["bucket_size_2p"]

        logging.info(
            "autotune_hp={}, train_iter={}, score={}".format(
                autotune_hp, train_iter, score
            )
        )
        record_autotune_log(self.autotune_log_filepath, autotune_hp, train_iter, score)

        tensor_list = [
            tensor_declar for bucket in bagua_hp.buckets for tensor_declar in bucket
        ]
        recommended_buckets = split_bucket_by_bucket_size(
            tensor_list,
            recommended_bucket_size,
            self.param_group_info,
        )
        logging.info(
            "bucket_size={}, recommended_bucket_size={}, recommended_buckets={}, is_hierarchical_reduce={}".format(
                2 ** bucket_size_2p,
                recommended_bucket_size,
                recommended_buckets,
                recommended_autotune_hp["is_hierarchical_reduce"],
            )
        )

        recommended_bagua_hp = BaguaHyperparameter(
            buckets=recommended_buckets,
            is_hierarchical_reduce=bool(
                recommended_autotune_hp["is_hierarchical_reduce"]
            ),
        )

        return recommended_bagua_hp, recommended_autotune_hp

    def setup_app(self, app):
        @app.route("/api/v1/bagua_backend_metrics", methods=["POST"])
        def report_bagua_backend_metrics():
            req: dict = request.get_json(force=True)  # TODO: @shjwudp
            tensor_ready_order: list = req["tensor_ready_order"]
            communication_time_ms: float = req["communication_time_ms"]
            hyperparameters: dict = req["hyperparameters"]

        @app.route("/api/v1/register_models", methods=["POST"])
        def register_models():
            req: dict = request.get_json(force=True)
            tensor_list: list[TensorDeclaration] = req["tensor_list"]
            param_group_info: Dict[str, int] = req["param_group_info"]
            whether_to_bucket: bool = req["whether_to_bucket"]

            self.sess_bucket_size = self.default_bucket_size
            if whether_to_bucket is False:
                self.sess_bucket_size = (
                    10 * 1024 ** 5
                )  # if you don't divide buckets, set big bucket as 10PB.
            session["bucket_size"] = self.sess_bucket_size

            with self.ask_hyperparameters_mutex:
                self.param_group_info = param_group_info
                default_hyperparameters = BaguaHyperparameter(
                    buckets=split_bucket_by_bucket_size(
                        tensor_list,
                        session["bucket_size"],
                        self.param_group_info,
                    ),
                )
                if self.is_initialized:
                    return json.dumps(
                        {
                            "message": "Autotune service has been initialized, and the current operation does not take effect.",
                            "recommended_hyperparameters": default_hyperparameters.dict(),
                        },
                    )

                self.recommended_hyperparameters = default_hyperparameters
                self.last_time_the_hyperparameters_was_granted = time.time()

                logging.info(
                    "tensor_list={}, buckets={}".format(
                        tensor_list, default_hyperparameters.buckets
                    )
                )
                self.is_initialized = True
                return json.dumps(
                    {
                        "recommended_hyperparameters": default_hyperparameters.dict(),
                    }
                )

        @app.route("/api/v1/report_metrics", methods=["POST"])
        def report_metrics():
            req: dict = request.get_json(force=True)
            rank: int = req["rank"]
            unix_timestamp: float = req["unix_timestamp"]
            iter_per_seconds: float = req["iter_per_seconds"]
            train_iter: int = req["train_iter"]
            denoised_iter_per_seconds: float = req["denoised_iter_per_seconds"]
            hyperparameters = req["hyperparameters"]

            if not self.is_initialized:
                return "Service not ready for ask_hyperparameters!", 405

            # Only consider the rank of the first report metrics now.
            with self.metrics_mutex:
                if train_iter <= rank0_train_iter.collect()[-1].samples[-1].value:
                    return json.dumps({})

                logging.info(
                    "rank={}, train_iter={}, denoised_iter_per_seconds={}, hyperparameters={}".format(
                        rank,
                        train_iter,
                        denoised_iter_per_seconds,
                        hyperparameters,
                    )
                )
                rank0_hyperparameters.info(hyperparameters)
                rank0_train_iter.set(train_iter)
                rank0_denoised_iter_per_seconds.set(denoised_iter_per_seconds)

            return json.dumps({})

        @app.route("/api/v1/checkboard", methods=["GET"])
        def request_checkboard():
            return json.dumps(
                {
                    "check_board": self.check_board,
                }
            )

        @app.route("/api/v1/ask_hyperparameters", methods=["POST"])
        def ask_hyperparameters():
            """
            report_metrics must be called before ask_hyperparameters
            """
            req: dict = request.get_json(force=True)
            rank: int = req["rank"]
            train_iter: int = req["train_iter"]

            if not self.is_initialized:
                return "Service not ready for ask_hyperparameters!", 405

            # # The bucket parameters requires at least one metrics report
            # while len(rank0_hyperparameters.collect()) == 0:
            #     time.sleep(0)

            def autotune():
                if self.sampling_counter > self.max_samples:
                    return

                with self.metrics_mutex:
                    recommended_train_iter = int(
                        rank0_train_iter.collect()[-1].samples[-1].value
                    )
                    denoised_iter_per_seconds = (
                        rank0_denoised_iter_per_seconds.collect()[-1].samples[-1].value
                    )
                    hyperparameters = (
                        rank0_hyperparameters.collect()[-1].samples[-1].labels
                    )
                    self.hyperparameters_and_score_list.append(
                        [
                            BaguaHyperparameter(**hyperparameters),
                            denoised_iter_per_seconds,
                        ]
                    )
                    hyperparameters_and_score_list = copy.deepcopy(
                        self.hyperparameters_and_score_list
                    )

                    logging.info(
                        "recommended_train_iter={}, hyperparameters={}, denoised_iter_per_seconds={}".format(
                            recommended_train_iter,
                            hyperparameters,
                            denoised_iter_per_seconds,
                        )
                    )

                # warmup pass
                if self.warmup_flag:
                    if (
                        time.time() - self.last_time_the_hyperparameters_was_granted
                        < self.warmup_time_s
                    ):
                        return
                    self.last_time_the_hyperparameters_was_granted = time.time()
                    self.warmup_flag = False

                # Skip if the sampling time is insufficient
                if (
                    time.time() - self.last_time_the_hyperparameters_was_granted
                    < self.sampling_confidence_time_s
                ):
                    return

                logging.info(
                    "rank={}, train_iter={}, sampling_counter={}, max_samples={}".format(
                        rank, train_iter, self.sampling_counter, self.max_samples
                    )
                )
                bagua_hp = BaguaHyperparameter(**hyperparameters)
                score = denoised_iter_per_seconds
                bucket_size_2p = int(math.log(session["bucket_size"], 2))
                (
                    recommended_bagua_hp,
                    recommended_autotune_hp,
                ) = self.tell_and_ask_bayesian_recommended_hyperparameters(
                    bagua_hp,
                    score,
                    recommended_train_iter,
                    bucket_size_2p,
                )

                if self.sampling_counter < self.max_samples:
                    self.recommended_hyperparameters = recommended_bagua_hp
                    self.recommended_from_iter = recommended_train_iter
                    self.sess_bucket_size = int(
                        2 ** recommended_autotune_hp["bucket_size_2p"]
                    )
                else:
                    # get best hyperparameters
                    sorted_score_hp = sorted(
                        [
                            (s, copy.deepcopy(p))
                            for p, s in hyperparameters_and_score_list
                        ],
                        key=lambda score_hp: score_hp[0],
                        reverse=True,
                    )
                    logging.info(
                        "sorted_score_hp={}".format(
                            [(s, p.dict()) for s, p in sorted_score_hp]
                        )
                    )
                    self.recommended_hyperparameters = sorted_score_hp[0][1]
                    self.recommended_from_iter = recommended_train_iter

                # The hyperparameters has been granted, update the time
                # NOTE: The accuracy depends on the client calling ask_hyperparameters at the right time and success
                self.last_time_the_hyperparameters_was_granted = time.time()
                self.sampling_counter += 1

            with self.ask_hyperparameters_mutex:
                # Autotune conditions:
                # 1. autotune_level >= 1.
                # 2. The bagua process is not in the process of hyperparameter update. (self.check_board.count(self.check_board[0])
                #   == len(self.check_board))
                # 3. Only execute autotune at most once in an iteration. (self.check_board[rank] < train_iter)
                if (
                    self.autotune_level >= 1
                    and self.check_board.count(self.check_board[0])
                    == len(self.check_board)
                    and self.check_board[rank] < train_iter
                ):
                    autotune()

                logging.debug(
                    "bucket_size={}, tarin_iter={}, rank={}".format(
                        session["bucket_size"], train_iter, rank
                    )
                )
                session["bucket_size"] = self.sess_bucket_size
                self.check_board[rank] = train_iter

                return json.dumps(
                    {
                        "recommended_hyperparameters": self.recommended_hyperparameters.dict(),
                        "recommended_from_iter": self.recommended_from_iter,
                        "is_autotune_processing": self.sampling_counter
                        < self.max_samples,
                    },
                )

        @app.route("/api/v1/reset", methods=["POST"])
        def reset():
            with self.ask_hyperparameters_mutex:
                if self.is_initialized:
                    self.__init__(
                        world_size=self.world_size,
                        autotune_level=self.autotune_level,
                        max_samples=self.max_samples,
                        sampling_confidence_time_s=self.sampling_confidence_time_s,
                    )

            return json.dumps({})

        # Add prometheus wsgi middleware to route /metrics requests
        app.wsgi_app = DispatcherMiddleware(app.wsgi_app, {"/metrics": make_wsgi_app()})

        # set secret-key
        app.config.update(SECRET_KEY=os.urandom(24))

        return app


class AutotuneClient:
    def __init__(self, service_addr: str, service_port: int):
        self.autotune_service_addr = "{}:{}".format(service_addr, service_port)
        self.session = requests.Session()

    def report_metrics(
        self,
        rank: int,
        unix_timestamp: float,
        train_iter: int,
        iter_per_seconds: float,
        denoised_iter_per_seconds: float,
        hyperparameters: dict,
        proxies={
            "http": None,
            "https": None,
        },
    ):
        try:
            rsp = self.session.post(
                "http://{}/api/v1/report_metrics".format(self.autotune_service_addr),
                json={
                    "rank": rank,
                    "unix_timestamp": unix_timestamp,
                    "train_iter": train_iter,
                    "iter_per_seconds": iter_per_seconds,
                    "denoised_iter_per_seconds": denoised_iter_per_seconds,
                    "hyperparameters": hyperparameters,
                },
                proxies=proxies,
            )
            return rsp.status_code
        except Exception as ex:
            logging.warning(
                "rank={}, train_iter={}, ex={}".format(rank, train_iter, ex)
            )
            return -1

    def register_models(
        self,
        tensor_list: List[TensorDeclaration],
        param_group_info: Dict[str, int] = {},
        whether_to_bucket: bool = True,
        proxies={
            "http": None,
            "https": None,
        },
    ):
        while True:
            try:
                rsp = self.session.post(
                    "http://{}/api/v1/register_models".format(
                        self.autotune_service_addr
                    ),
                    json={
                        "tensor_list": tensor_list,
                        "param_group_info": param_group_info,
                        "whether_to_bucket": whether_to_bucket,
                    },
                    proxies=proxies,
                )
                assert (
                    rsp.status_code == 200
                ), "request register_models failed, rsp={}".format(rsp)
                return rsp
            except Exception as ex:
                logging.warning("ex={}".format(ex))
                time.sleep(1)
                continue

    def request_checkboard(self):
        while True:
            try:
                rsp = self.session.get(
                    "http://{}/api/v1/checkboard".format(self.autotune_service_addr)
                )
                return rsp
            except Exception as ex:
                logging.warning("ex={}".format(ex))
                time.sleep(1)
                continue

    def wait_for_all_process_parameters_updated(self, my_train_iter: int):
        while True:
            rsp = self.request_checkboard()
            assert (
                rsp.status_code == 200
            ), "rsp={}, 503 may be due to http_proxy".format(rsp)
            check_board = rsp.json()["check_board"]
            if all([train_iter == my_train_iter for train_iter in check_board]):
                break
            logging.info("check_board={}".format(check_board))
            time.sleep(1)

    def ask_hyperparameters(
        self,
        rank: int,
        train_iter: int,
        proxies={
            "http": None,
            "https": None,
        },
    ):
        while True:
            try:
                rsp = self.session.post(
                    "http://{}/api/v1/ask_hyperparameters".format(
                        self.autotune_service_addr
                    ),
                    json={
                        "rank": rank,
                        "train_iter": train_iter,
                    },
                    proxies=proxies,
                )
                assert rsp.status_code == 200, "rsp={}".format(rsp)
                return rsp
            except Exception as ex:
                logging.warning("rank={}, exception={}".format(rank, ex))
                time.sleep(1)
                continue

    def reset(
        self,
        proxies={
            "http": None,
            "https": None,
        },
    ):
        while True:
            try:
                rsp = self.session.post(
                    "http://{}/api/v1/reset".format(self.autotune_service_addr),
                    proxies=proxies,
                )
                return rsp
            except Exception as ex:
                logging.warning("ex={}".format(ex))
                time.sleep(1)
                continue


if __name__ == "__main__":
    autotune_service = AutotuneService(10)
    app = Flask(__name__)
    app = autotune_service.setup_app(app)

    server = multiprocessing.Process(
        target=app.run,
        kwargs={
            "host": "0.0.0.0",
            "port": 8123,
        },
    )
    server.daemon = True
    server.start()
    server.join()



import unittest
import logging
import requests
import time
import multiprocessing
import numpy as np
from flask import Flask
from bagua.autotune import AutoTuneHyperParams
from bagua.service import AutotuneService, BaguaHyperparameter, pick_n_free_ports


def metrics(buckets, is_hierarchical_reduce):
    # score = sum(-abs(bucket_sum_size - 5M))
    # 峰在 bucket_size=5M 的凸函数
    score = 0.0
    for bucket in buckets:
        score += -abs(sum([td["num_elements"] for td in bucket]) - 5 * 1024 ** 2)

    if not is_hierarchical_reduce:
        score += abs(score) * 0.1

    return score


class AutotuneClient:
    def __init__(self, autotune_service_addr):
        self.autotune_service_addr = autotune_service_addr

    def report_metrics(
        self,
        rank,
        unix_timestamp,
        train_iter,
        iter_per_seconds,
        denoised_iter_per_seconds,
        hyperparameters,
    ):
        try:
            rsp = requests.post(
                "http://{}/api/v1/report_metrics".format(self.autotune_service_addr),
                json={
                    "rank": rank,
                    "unix_timestamp": unix_timestamp,
                    "train_iter": train_iter,
                    "iter_per_seconds": iter_per_seconds,
                    "denoised_iter_per_seconds": denoised_iter_per_seconds,
                    "hyperparameters": hyperparameters,
                },
            )
            return rsp.status_code
        except Exception as ex:
            logging.warning(
                "rank={}, train_iter={}, ex={}".format(rank, train_iter, ex)
            )
            return -1

    def request_checkboard(self):
        while True:
            try:
                rsp = requests.get(
                    "http://{}/api/v1/checkboard".format(self.autotune_service_addr)
                )
                return rsp
            except Exception as ex:
                logging.warning("ex={}".format(ex))
                continue

    def wait_for_all_process_parameters_updated(self, my_train_iter):
        while True:
            rsp = self.request_checkboard()
            assert rsp.status_code == 200, "rsp={}".format(rsp)
            check_board = rsp.json()["check_board"]
            if all([train_iter >= my_train_iter for train_iter in check_board]):
                break
            time.sleep(0)

    def ask_hyperparameters(self, rank, train_iter):
        while True:
            try:
                rsp = requests.post(
                    "http://{}/api/v1/ask_hyperparameters".format(
                        self.autotune_service_addr
                    ),
                    json={
                        "rank": rank,
                        "train_iter": train_iter,
                    },
                )
                return rsp
            except Exception as ex:
                logging.warning("rank={}, exception={}".format(rank, ex))
                time.sleep(1)
                continue


class MockBaguaProcess:
    def __init__(self, rank, autotune_service_addr):
        self.rank = rank
        self.iter_per_seconds = 0.0
        self.denoised_iter_per_seconds = 0.0
        self.hyperparameters = BaguaHyperparameter(
            buckets=[
                [
                    {
                        "name": "A",
                        "num_elements": 1 * 1024 ** 2,
                        "dtype": "F32",
                    }
                ],
                [
                    {
                        "name": "B",
                        "num_elements": 3 * 1024 ** 2,
                        "dtype": "F32",
                    }
                ],
                [
                    {
                        "name": "C",
                        "num_elements": 5 * 1024 ** 2,
                        "dtype": "F32",
                    }
                ],
                [
                    {
                        "name": "D",
                        "num_elements": 7 * 1024 ** 2,
                        "dtype": "F32",
                    }
                ],
                [
                    {
                        "name": "E",
                        "num_elements": 11 * 1024 ** 2,
                        "dtype": "F32",
                    }
                ],
            ],
            is_hierarchical_reduce=False,
        )
        self.client = AutotuneClient(autotune_service_addr)

    def run(self, total_iters=1200):
        for i in range(total_iters):
            if i != 0 and i % 10 == 0:
                rsp = self.client.ask_hyperparameters(self.rank, i)
                assert rsp.status_code == 200, "rsp={}".format(rsp)
                self.hyperparameters.update(rsp.json()["recommended_hyperparameters"])
                self.client.wait_for_all_process_parameters_updated(i)

            score = metrics(
                self.hyperparameters.buckets,
                self.hyperparameters.is_hierarchical_reduce,
            )
            logging.info(
                "train_iter={}, score={}, hyperparameters={}".format(
                    i, score, self.hyperparameters.dict()
                )
            )
            ret = self.client.report_metrics(
                self.rank, time.time(), i, score, score, self.hyperparameters.dict()
            )
            assert ret == 200, "ret={}".format(ret)

        score = metrics(
            self.hyperparameters.buckets,
            self.hyperparameters.is_hierarchical_reduce,
        )
        logging.info(
            "best score={}, best hyperparameters={}".format(
                score, self.hyperparameters.dict()
            )
        )
        assert score == -8493465.6, "score={}".format(score)

        return score


class TestAutotuneService(unittest.TestCase):
    def test_autotune_service(self):
        master_addr = "127.0.0.1"
        master_port = pick_n_free_ports(1)[0]
        autotune_service_addr = "{}:{}".format(master_addr, master_port)
        nprocs = 2

        autotune_service = AutotuneService(
            nprocs, max_samples=100, sampling_confidence_time_s=0
        )
        app = Flask(__name__)
        app = autotune_service.setup_app(app)

        server = multiprocessing.Process(
            target=app.run,
            kwargs={
                "host": "0.0.0.0",
                "port": master_port,
                "debug": False,
            },
        )
        server.daemon = True
        server.start()

        mock_objs = []
        with multiprocessing.pool.ThreadPool(nprocs) as pool:
            results = []
            for i in range(nprocs):
                mock = MockBaguaProcess(i, autotune_service_addr)
                mock_objs.append(mock)
                ret = pool.apply_async(mock.run)
                results.append(ret)
            for ret in results:
                score = ret.get()
                self.assertEqual(score, -8493465.6)

        server.terminate()
        server.join()


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(filename)s: %(levelname)s: %(funcName)s(): %(lineno)d:\t%(message)s",
    )
    unittest.main()



#!/usr/bin/env python3

from bagua.torch_api.bucket import BaguaBucket
from bagua.torch_api.distributed import BaguaModule
from bagua.torch_api.algorithms import Algorithm


class GradientAllReduceAlgorithm(Algorithm):
    def __init__(self, hierarchical: bool = False, average: bool = True):
        """
        Create an instance of the
        `GradientAllReduce <https://baguasys.github.io/tutorials/algorithms/gradient-allreduce.html>`_
        algorithm.

        Args:
            hierarchical (bool): Enable hierarchical communication.
            average (bool): If True, the gradients on each worker are averaged.
                Otherwise, they are summed.
        """
        self.hierarchical = hierarchical
        self.average = average

    def init_operations(
        self,
        bagua_module: BaguaModule,
        bucket: BaguaBucket,
    ):
        bucket.clear_ops()
        if self.hierarchical:
            bucket.append_centralized_synchronous_op(
                hierarchical=self.hierarchical,
                average=self.average,
            )
        else:
            bucket.append_centralized_synchronous_op(
                hierarchical=self.hierarchical,
                average=self.average,
            )



import unittest
from bagua.service import pick_n_free_ports, generate_and_broadcast_server_addr


class TestServiceDiscovery(unittest.TestCase):
    def test_generate_and_broadcast_server_addr(self):
        master_addr = "127.0.0.1"
        master_port = pick_n_free_ports(1)[0]

        import multiprocessing

        n_agent = 3
        result = list(range(n_agent))

        def get_addr(i):
            result[i] = generate_and_broadcast_server_addr(
                master_addr, master_port, n_agent, i
            )

        jobs = []
        for i in range(n_agent):
            p = multiprocessing.Process(target=get_addr, args=(i,))
            p.start()
            jobs.append(p)

        for p in jobs:
            p.join()

        for x in result:
            self.assertEqual(x, result[0])


if __name__ == "__main__":
    unittest.main()



from bagua.torch_api.distributed import init_communicator
import cupy
import torch
import argparse
import os
import logging

"""
run the script with the following commands:
python3 -m torch.distributed.launch --nproc_per_node=8 p2p.py
"""
logging.basicConfig(level=logging.DEBUG)
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--local_rank", default=0, type=int)
    parser.add_argument(
        "--rank", default=-1, type=int, help="node rank for distributed training"
    )

    args = parser.parse_args()
    # need to initialize torch.dist
    torch.cuda.set_device(args.local_rank)
    torch.distributed.init_process_group(backend="nccl", init_method="env://")

    comm = init_communicator("nccl")

    torch.cuda.set_device(comm.get_rank())

    print("rank: {}, size: {}".format(comm.get_rank(), comm.get_size()))

    sendbuf = cupy.random.random(10)
    recvbuf = cupy.zeros([comm.get_size(), sendbuf.size], dtype=sendbuf.dtype)

    sendbuf2 = cupy.arange(10)
    recvbuf2 = cupy.zeros([comm.get_size(), sendbuf2.size], dtype=sendbuf2.dtype)

    comm.gather(sendbuf, recvbuf, 0)
    comm.gather(sendbuf2, recvbuf2, 1)

    print("{} after send recv, {}, {}".format(comm.get_rank(), recvbuf, recvbuf2))



import logging
import multiprocessing
import bagua_core as B
from flask import Flask
import bagua.torch_api.globals
from bagua.service import AutotuneService
from .env import (
    get_world_size,
    get_rank,
    get_local_rank,
    get_local_size,
    get_master_addr,
    get_default_bucket_size,
    get_bagua_service_port,
)
from .globals import _get_global_state, is_initialized
from ..service.autotune_service import AutotuneClient
from .exceptions import RepeatedInitializationError
from .utils import flatten, unflatten, to_bagua_reduce_op
from ..bagua_define import BaguaHyperparameter
import torch
import torch.distributed as dist
import torch.distributed.distributed_c10d as c10d

_autotune_server = None


def start_autotune_server():
    """Start autotune server in background."""
    global _autotune_server

    autotune_service = AutotuneService(
        world_size=get_world_size(),
        default_bucket_size=get_default_bucket_size(),
    )
    app = Flask(__name__)
    app = autotune_service.setup_app(app)
    log = logging.getLogger("werkzeug")
    log.setLevel(logging.ERROR)

    _autotune_server = multiprocessing.Process(
        target=app.run,
        kwargs={
            "host": "0.0.0.0",
            "port": get_bagua_service_port(),
            "debug": False,
        },
    )
    _autotune_server.daemon = True
    _autotune_server.start()


def init_process_group():
    """Initializes the PyTorch builtin distributed process group, and this will
    also initialize the distributed package, should be executed before all the
    APIs of bagua.

    Raises:
        RepeatedInitializationError: If you run this function repeatedly

    Examples::
        >>> import bagua.torch_api as bagua
        >>> bagua.init_process_group()
        >>> model = torch.nn.Sequential(
        ...    torch.nn.Linear(D_in, H),
        ...    torch.nn.ReLU(),
        ...    torch.nn.Linear(H, D_out),
        ...    )
        >>> optimizer = torch.optim.SGD(
        ...    model.parameters(),
        ...    lr=0.01,
        ...    momentum=0.9
        ...    )
        >>> model, optimizer = bagua_init(model, optimizer)
    """
    if is_initialized():
        raise RepeatedInitializationError()

    if not dist.is_initialized():
        torch.distributed.init_process_group(
            backend="nccl", init_method="env://"
        )  # fmt: off

    store = c10d._get_default_store()

    if get_rank() == 0:
        start_autotune_server()

    bagua.torch_api.globals._set_global_state(BaguaGlobalState(store))


class BaguaGlobalState(object):
    def __init__(self, store=None, device_id=None):
        if device_id is None:
            device_id = get_local_rank()
        self.backend = B.BaguaCommBackendPy(100, device_id=device_id)
        self.stream = torch.cuda.Stream(priority=-1)
        self.store = store
        self.hyperparameters = BaguaHyperparameter()
        self.hyperparameters_service_client = AutotuneClient(
            get_master_addr(), get_bagua_service_port()
        )
        self.internode_communicator = init_bagua_inter_communicator(
            stream=self.stream, leader_rank=0, store=self.store, device_id=device_id
        )
        self.intranode_communicator = init_bagua_intra_communicator(
            stream=self.stream, store=self.store, device_id=device_id
        )
        self.global_communicator = init_bagua_communicator(
            stream=self.stream, store=self.store, device_id=device_id
        )

    def get_communication_stream(self):
        return self.stream

    def get_internode_communicator(self):
        return self.internode_communicator

    def get_intranode_communicator(self):
        return self.intranode_communicator

    def get_global_communicator(self):
        return self.global_communicator

    def get_backend(self):
        return self.backend


def get_bagua_hyperparameters():
    return _get_global_state().hyperparameters


def get_hyperparameters_service_client():
    return _get_global_state().hyperparameters_service_client


def gen_nccl_unique_id(comm_type: str, root=0, store=None):
    key = f"{comm_type}-{root}-unique_id"

    if store is None:
        store = c10d._get_default_store()

    if get_rank() == root:
        idstr = B.BaguaSingleCommunicatorPy.generate_nccl_unique_id_str()
        store.set(key, idstr)
    else:
        idstr = store.get(key)
        idstr = str(idstr, encoding="utf-8")

    return idstr


def init_bagua_inter_communicator(stream, leader_rank=0, store=None, device_id=None):
    if device_id is None:
        device_id = get_local_rank()
    nccl_unique_id = gen_nccl_unique_id(
        "bagua_inter_comm", root=leader_rank, store=store
    )

    if get_rank() % get_local_size() != leader_rank:
        return None

    comm = B.BaguaSingleCommunicatorPy(
        rank=get_rank() // get_local_size(),
        nranks=get_world_size() // get_local_size(),
        device_id=device_id,
        stream_ptr=stream.cuda_stream,
        nccl_unique_id_str=nccl_unique_id,
    )
    comm.cuda_stream = stream
    logging.debug(
        "init bagua internode communicator ok, global rank: %s rank: %s",
        dist.get_rank(),
        comm.rank(),
    )
    return comm


def init_bagua_intra_communicator(stream, store=None, device_id=None):
    if device_id is None:
        device_id = get_local_rank()
    nccl_unique_id = gen_nccl_unique_id(
        "bagua_intra_comm",
        root=get_rank() // get_local_size() * get_local_size(),
        store=store,
    )

    comm = B.BaguaSingleCommunicatorPy(
        rank=get_rank() % get_local_size(),
        nranks=get_local_size(),
        device_id=device_id,
        stream_ptr=stream.cuda_stream,
        nccl_unique_id_str=nccl_unique_id,
    )
    comm.cuda_stream = stream
    logging.debug(
        "init bagua intranode communicator ok, global rank: %s rank: %s",
        dist.get_rank(),
        comm.rank(),
    )
    return comm


def init_bagua_communicator(stream, store=None, device_id=None):
    if device_id is None:
        device_id = get_local_rank()
    nccl_unique_id = gen_nccl_unique_id("bagua_global_comm", store=store)

    comm = B.BaguaSingleCommunicatorPy(
        rank=get_rank(),
        nranks=get_world_size(),
        device_id=device_id,
        stream_ptr=stream.cuda_stream,
        nccl_unique_id_str=nccl_unique_id,
    )
    comm.cuda_stream = stream
    logging.debug(
        "init bagua global communicator ok, global rank: %s rank: %s",
        dist.get_rank(),
        comm.rank(),
    )
    return comm


def broadcast_coalesced(tensors, root=0, comm: B.BaguaSingleCommunicatorPy = None):
    for tensor in tensors:
        assert tensor.device != torch.device(
            "cpu"
        ), "input tensors must be CUDA and dense"

    if comm is None:
        comm = _get_global_state().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        coalesced = flatten(tensors)
        comm.broadcast(coalesced.to_bagua_tensor().bagua_backend_tensor(), root)
        for buf, synced in zip(tensors, unflatten(coalesced, tensors)):
            buf.copy_(synced)

    # TODO: remove
    torch.cuda.synchronize()


def broadcast(tensor, root=0, comm: B.BaguaSingleCommunicatorPy = None):
    r"""Broadcasts the tensor to the whole communicator.

    `tensor` must have the same number of elements in all processes
    participating in the collective.

    Args:
        tensor (torch.Tensor): Data to be sent if `root` is the rank of
            current process, and tensor to be used to save received data
            otherwise.
        root (int, optional): Source rank. Defaults to 0.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator
            to work on. If None, the global bagua communicator will be used.
            Defaults to None.
    """  # noqa: W293

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None:
        comm = _get_global_state().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.broadcast(tensor.to_bagua_tensor().bagua_backend_tensor(), root)

    # TODO: remove
    torch.cuda.synchronize()


def reduce(tensor, dst, op=dist.ReduceOp.SUM, comm: B.BaguaSingleCommunicatorPy = None):
    r"""Reduces the tensor across all processes.

    Only the process whit rank `dst` is going to receive the final result.

    Args:
        tensor (torch.Tensor): Input and output of the collective. The
            function operates in-place.
        dst (int): Destination rank
        op (optional): one of the values from `torch.distributed.ReduceOp`
            enum. Specifies an operation used for element-wise reductions.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator to
            work on. If None the global bagua communicator will be used.
            Defaults to None.
    """  # noqa: W293

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None:
        comm = _get_global_state().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.reduce(
            tensor.to_bagua_tensor().bagua_backend_tensor(), dst, to_bagua_reduce_op(op)
        )

    torch.cuda.synchronize()


def allreduce_coalesced(
    tensors,
    op=dist.ReduceOp.SUM,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    for tensor in tensors:
        assert tensor.device != torch.device(
            "cpu"
        ), "input tensors must be CUDA and dense"

    if comm is None:
        comm = _get_global_state().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        coalesced = flatten(tensors)
        comm.allreduce(
            coalesced.to_bagua_tensor("allreduce_coalesced"), to_bagua_reduce_op(op)
        )

        for buf, synced in zip(tensors, unflatten(coalesced, tensors)):
            buf.copy_(synced)

    # TODO: remove
    torch.cuda.synchronize()


def allreduce(
    tensor,
    op=dist.ReduceOp.SUM,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """Reduces the tensor data across all machines in such a way that all get
    the final result. After the call tensor is going to be bitwise identical
    in all processes.

    Args:
        tensor (torch.Tensor): Input and output of the collective. The
            function operates in-place.
        op (optional): one of the values from `torch.distributed.ReduceOp` enum. Specifies an operation used for element-wise reductions.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator to
            work on. If None the global bagua communicator will be used.
            Defaults to None.

    Examples:
        >>> from bagua.torch_api import allreduce
        >>> # All tensors below are of torch.int64 type.
        >>> # We have 2 process groups, 2 ranks.
        >>> tensor = torch.arange(2, dtype=torch.int64) + 1 + 2 * rank
        >>> tensor
        tensor([1, 2]) # Rank 0
        tensor([3, 4]) # Rank 1
        >>> allreduce(tensor)
        >>> tensor
        tensor([4, 6]) # Rank 0
        tensor([4, 6]) # Rank 1

        >>> # All tensors below are of torch.cfloat type.
        >>> # We have 2 process groups, 2 ranks.
        >>> tensor = torch.tensor([1+1j, 2+2j], dtype=torch.cfloat) + 2 * rank * (1+1j)
        >>> tensor
        tensor([1.+1.j, 2.+2.j]) # Rank 0
        tensor([3.+3.j, 4.+4.j]) # Rank 1
        >>> allreduce(tensor)
        >>> tensor
        tensor([4.+4.j, 6.+6.j]) # Rank 0
        tensor([4.+4.j, 6.+6.j]) # Rank 1
    """  # noqa: E501

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None:
        comm = _get_global_state().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.allreduce(
            tensor.to_bagua_tensor().bagua_backend_tensor(), to_bagua_reduce_op(op)
        )

    # TODO: remove
    torch.cuda.synchronize()



import os
from distutils.errors import (
    DistutilsPlatformError,
)
from setuptools import setup, find_packages
import sys


def check_torch_version():
    try:
        import torch
    except ImportError:
        print("import torch failed, is it installed?")

    version = torch.__version__
    if version is None:
        raise DistutilsPlatformError(
            "Unable to determine PyTorch version from the version string '%s'"
            % torch.__version__
        )
    return version


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))

    setup(
        name="bagua",
        use_scm_version={"local_scheme": "no-local-version"},
        setup_requires=["setuptools_scm"],
        url="https://github.com/BaguaSys/bagua",
        python_requires=">=3.7",
        description="Bagua is a flexible and performant distributed training algorithm development framework.",
        packages=find_packages(exclude=("tests")),
        author="Kuaishou AI Platform & DS3 Lab",
        author_email="admin@mail.xrlian.com",
        install_requires=[
            "bagua-core>=0.3,<0.4",
            "deprecation",
            "pytest-benchmark",
            "scikit-optimize",
            "numpy",
            "flask",
            "prometheus_client",
            "parallel-ssh",
            "pydantic",
            "requests",
            "gorilla",
        ],
        entry_points={
            "console_scripts": [
                "baguarun = bagua.script.baguarun:main",
            ],
        },
        scripts=[
            "bagua/script/bagua_sys_perf",
        ],
    )



import bagua.torch_api as bagua
import torch
import logging
import argparse


logging.basicConfig(level=logging.DEBUG)
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--local_rank", default=0, type=int)
    parser.add_argument(
        "--rank", default=0, type=int, help="node rank for distributed training"
    )

    args = parser.parse_args()

    # need to initialize torch.dist
    torch.cuda.set_device(args.local_rank)
    bagua.init_process_group()

    if args.rank == 0:
        tensors = [torch.rand(10).cuda() for _ in range(3)]
    else:
        tensors = [torch.zeros(10).cuda() for _ in range(3)]

    bagua.broadcast_coalesced(tensors, root=0)

    print("{}, after bcast, {}".format(bagua.get_rank(), tensors))

    tensor = torch.rand(10).cuda()
    bagua.allreduce(tensor)

    print("{}, after bcast, {}".format(bagua.get_rank(), tensor))
